/* Written by Dmitrij Schlesinger (2012),
 * www.inf.tu-dresden.de/~ds24/rtsegm/rtsegm.html
 *
 * The two classes below are a minimalistic wrapper for
 * the necessary OpenCV functionality. Tested with
 * OpenCV-2.3.1 under Ubuntu Linux 12.04 (gcc 4.6.3)
 * WindowsXP and Windows7 (Microsoft Visual Studio 2010).
*/

#include "d_inout.h"

d_in::d_in(int mode, int camera, char *filename, int height, int width, bool flip) {
    cap=0;
    if( mode==0 || mode==1 ) { // video from webcam or file
        // open
        if( mode ) cap = new VideoCapture(filename); // file
        else cap = new VideoCapture(camera);
        if( !cap || !cap->isOpened() ) return;

        // resizing -- define "manual_scale"
        if( height && width ) { // sizes are relevant -- care about scaling
            // look for actual sizes (do not touch global frames)
            Mat frame_tmp;
            *cap >> frame_tmp;
            if( frame_tmp.rows==height &&
                    frame_tmp.cols==width ) manual_scale=false; // not necessary to scale
            else { // force scaling
                cap->set(CV_CAP_PROP_FRAME_HEIGHT,(double)height);
                cap->set(CV_CAP_PROP_FRAME_WIDTH,(double)width);
                // re-read
                frame_tmp.release();
                *cap >> frame_tmp;
                if( frame_tmp.rows==height &&
                        frame_tmp.cols==width ) manual_scale=false; // success
                else manual_scale=true; // did not work -- should be rescaled manually
            } // end of forcing
            frame_tmp.release();
        } // end of "sizes are relevant"
        else manual_scale=false; // sizes are not relevant

        *cap >> frame1;
        if( flip ) {
            if( manual_scale ) {
                // cam->(resize)->cam1->(flip)->frame
                resize(frame1,frame2,Size(width,height));
                cv::flip(frame2,frame3,1);
                img=frame3.ptr(0);
                this->height=frame3.rows;
                this->width=frame3.cols;
            } else {
                // cam->(flip)->frame
                cv::flip(frame1,frame2,1);
                img=frame2.ptr(0);
                this->height=frame2.rows;
                this->width=frame2.cols;
            }
        } else {
            if( manual_scale ) {
                // cam->(resize)->frame
                resize(frame1,frame2,Size(width,height));
                img=frame2.ptr(0);
                this->height=frame2.rows;
                this->width=frame2.cols;
            } else {
                // frame
                img=frame1.ptr(0);
                this->height=frame1.rows;
                this->width=frame1.cols;
            }
        }
    } // end of "camera or video;
    else if( mode==2 ) { // image
        frame1=imread(filename);
        int height_tmp=frame1.rows;
        int width_tmp=frame1.cols;

        // care about scale
        if( height && width && (height!=height_tmp || width!=width_tmp) )
            resize(frame1,frame1,Size(width,height));
        if( flip ) cv::flip(frame1,frame1,1);
        img=frame1.ptr(0);
        this->height=frame1.rows;
        this->width=frame1.cols;
    } else return; // unknown input mode

    // remember all
    this->mode=mode;
    this->flip=flip;
}

d_in::~d_in() {
    frame1.release();
    frame2.release();
    frame3.release();
    if( mode==0 || mode==1 ) {
        if( cap ) delete cap;
    }
}

bool d_in::in() {
    if( mode>1 ) return true;
    if( !cap->grab() ) return false;
    cap->retrieve(frame1);
    if( flip ) {
        if( manual_scale ) {
            // cam->(resize)->cam1->(flip)->frame
            resize(frame1,frame2,Size(width,height));
            cv::flip(frame2,frame3,1);
        } else {
            // cam->(flip)->frame
            cv::flip(frame1,frame2,1);
        }
    } else {
        if( manual_scale ) {
            // cam->(resize)->frame
            resize(frame1,frame2,Size(width,height));
        }
    }
    return true;
}

int d_in::key(int delay) {
    return(waitKey(delay));
}

d_out::d_out(int mode, char *name, int height, int width, unsigned char *img) {
    frame=new Mat(height,width,CV_8UC3,img);
    switch( mode ) {
    case 0: // window
        namedWindow(name,CV_WINDOW_NORMAL|CV_WINDOW_KEEPRATIO|CV_GUI_EXPANDED);
        break;
    case 1: // video
        video = new VideoWriter(name,CV_FOURCC('D','I','V','X'),30.,frame->size(),true);
        if( !video || !video->isOpened() ) {
            if( video ) delete video;
            video=0;
        }
        break;
    case 2: // image
        break;
    default: // error
        break;
    }

    this->mode=mode;
    this->name=name;
    this->height=height;
    this->width=width;
}

d_out::~d_out() {
    switch( mode ) {
    case 0:
        destroyWindow(name);
        break;
    case 1:
        if( video ) delete video;
        break;
    case 2:
    default:
        break;
    }

    delete frame;
}

void d_out::out() {
    switch( mode ) {
    case 0:
        imshow(name,*frame);
        break;
    case 1:
        if( video && video->isOpened() ) *video << *frame;
        break;
    case 2:
        imwrite(name,*frame);
        break;
    default:
        break;
    }
}

void d_out::draw_ellipse(int x0, int y0, int rx, int ry) {
    ellipse(*frame,Point(x0,y0),Size(rx,ry),0.,0.,360.,Scalar(0,0,255));
}
