/*
 * Written by Eeri Kask ...--2012
 *
 * a minimalistic wrapper for threading management
 * This is a mapping to "POSIX threads" under MAC/Linux
 * and "Win32-threads" under Windows
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "ed_thread.h"

#if defined (_WIN32) || defined (_WIN64)
// empty
#elif defined (__MACH__) && defined (__APPLE__)
#include <mach/mach.h>
#include <mach/machine.h>
#else
#include <unistd.h>
#endif

int ed_thread::num_threads() {
    int corez = 0;

#if defined (_WIN32) || defined (_WIN64)
    SYSTEM_INFO sinf;
    GetSystemInfo (&sinf);
    corez = sinf.dwNumberOfProcessors;
#elif defined (__MACH__) && defined (__APPLE__)
    struct host_basic_info hinf;
    unsigned int cnt = HOST_BASIC_INFO_COUNT;
    if (host_info (mach_host_self(), HOST_BASIC_INFO, (host_info_t)(&hinf), &cnt) == KERN_SUCCESS)
        corez = hinf.avail_cpus;  // max_cpus avail_cpus physical_cpu_max physical_cpu logical_cpu_max logical_cpu
#else
    corez = sysconf (_SC_NPROCESSORS_CONF);
#endif
    if (corez < 1) corez = 1;
    return corez;
}

#if defined (_WIN32) || defined (_WIN64)

static DWORD WINAPI Circular (void *pointer) {
    // entering here as user-thread
    ed_thread *edth=(ed_thread *)pointer;
    for (;;) {
        switch (WaitForMultipleObjects (3-1, (edth->msg)+1, FALSE, INFINITE)) {
            // wait for 'run' or 'terminate' from main-thread
        default:
            //      case WAIT_OBJECT_0 + 0 - 1:
            continue;
        case WAIT_OBJECT_0 + 1 - 1:
            // 'run' received from main-thread
            // one-time execute user-provided function
            edth->run();
            SetEvent (edth->msg[0]); // send 'wait' (i.e. 'completed') to main-thread
            continue;
        case WAIT_OBJECT_0 + 2 - 1: // 'terminate' received from main-thread, break-out
            break;
        case WAIT_FAILED:
            fprintf (stderr, "Circular: return code from WaitForMultipleObjects() is WAIT_FAILED\n");
            ExitThread (1);
        }
        break;
    }
    return 0; // user-thread termination
}

void ed_thread::start () {
    // entering here as main-thread
    SetEvent (msg[1]); // send 'run' to user-thread
}

void ed_thread::wait () {
    // entering here as main-thread
    WaitForSingleObject (msg[0], INFINITE);  // wait for 'wait' (i.e. 'completed') from user-thread
}

ed_thread::ed_thread () {
    // entering here as main-thread
    msg[0] = CreateEvent (/*sec*/NULL, /*manual-reset*/FALSE, /*initially-signaled*/FALSE, NULL);
    if (msg[0] == NULL) fprintf (stderr, "ed_thread: return code from CreateEvent(0) is NULL\n");
    msg[1] = CreateEvent (/*sec*/NULL, /*manual-reset*/FALSE, /*initially-signaled*/FALSE, NULL);
    if (msg[1] == NULL) fprintf (stderr, "ed_thread: return code from CreateEvent(1) is NULL\n");
    msg[2] = CreateEvent (/*sec*/NULL, /*manual-reset*/FALSE, /*initially-signaled*/FALSE, NULL);
    if (msg[2] == NULL) fprintf (stderr, "ed_thread: return code from CreateEvent(2) is NULL\n");
    thrhnd = CreateThread (/*sec*/NULL, /*stack*/0, Circular, this, /*flags*/0, &id);
    if (thrhnd == NULL) fprintf (stderr, "ed_thread: return code from CreateThread() is NULL\n");
}

ed_thread::~ed_thread() {
    // entering here as main-thread
    SetEvent (msg[2]); // send 'terminate' to user-thread
    WaitForSingleObject (thrhnd, INFINITE); // wait here for 'thread termination' of user-thread
    CloseHandle (thrhnd);
    CloseHandle (msg[0]);
    CloseHandle (msg[1]);
    CloseHandle (msg[2]);
}

#else /*PTHREADS*/

void *Circular(void *pointer) {
    // entering here as user-thread
    ed_thread *edth=(ed_thread *)pointer;
    for (;;) {
        int rc;
        rc = pthread_mutex_lock (&edth->mutx_run);
        if (rc) fprintf (stderr, "Circular: return code from pthread_mutex_lock(&mutx_run) is %d\n", rc);

        while (edth->stat_run == 0) { // sleep as long as we (user-thread) are asked to stay in 'wait' state
            rc = pthread_cond_wait (&edth->cond_run,&edth->mutx_run);
            if (rc) fprintf (stderr, "Circular: return code from pthread_cond_wait(&cond_run) is %d\n", rc);
        }

        if (edth->stat_run == (unsigned)(~0)) {
            // ('2') 'terminate' received from main-thread, break-out
            rc = pthread_mutex_unlock (&edth->mutx_run);
            if (rc) fprintf (stderr, "Circular: return code from pthread_mutex_unlock(&mutx_run) is %d\n", rc);
            break;
        } else { // if (thrcir_data_p->run == 1)
            // 'run' received from main-thread
            edth->run();
        }

        edth->stat_run = 0; // we (user-thread) go into 'wait' (i.e. 'completed') state

        rc = pthread_cond_signal (&edth->cond_run); // send 'wait' (i.e. 'completed') to main-thread
        if (rc) fprintf (stderr, "Circular: return code from pthread_cond_signal(&cond_run) is %d\n", rc);

        rc = pthread_mutex_unlock (&edth->mutx_run);
        if (rc) fprintf (stderr, "Circular: return code from pthread_mutex_unlock(&mutx_run) is %d\n", rc);
    }
    return (void*)(0); // user-thread termination
}

void ed_thread::start() {
    // entering here as main-thread
    int rc;
    rc = pthread_mutex_lock (&mutx_run);
    if (rc) fprintf (stderr, "start: return code from pthread_mutex_lock(&mutx_run) is %d\n", rc);

    while (stat_run != 0) { // sleep as long as we (main-thread) haven't got 'wait' (i.e. 'completed') from user-thread
        rc = pthread_cond_wait (&cond_run,&mutx_run);
        if (rc) fprintf (stderr, "start: return code from pthread_cond_wait(&cond_run) is %d\n", rc);
    }

    stat_run = 1; // send 'run' to user-thread (any value not '0' and not '(~0)' is OK)

    rc = pthread_cond_signal (&cond_run);
    if (rc) fprintf (stderr, "start: return code from pthread_cond_signal(&cond_run) is %d\n", rc);

    rc = pthread_mutex_unlock (&mutx_run);
    if (rc) fprintf (stderr, "start(): return code from pthread_mutex_unlock(&mutx_run) is %d\n", rc);
}

void ed_thread::wait() {
    // entering here as main-thread
    int rc;
    rc = pthread_mutex_lock (&mutx_run);
    if (rc) fprintf (stderr, "stop: return code from pthread_mutex_lock(&mutx_run) is %d\n", rc);

    while (stat_run != 0) { // sleep as long as we (main-thread) haven't got 'wait' (i.e. 'completed') from user-thread
        rc = pthread_cond_wait (&cond_run,&mutx_run);
        if (rc) fprintf (stderr, "stop: return code from pthread_cond_wait(&cond_run) is %d\n", rc);
    }

    rc = pthread_mutex_unlock (&mutx_run);
    if (rc) fprintf (stderr, "stop: return code from pthread_mutex_unlock(&mutx_run) is %d\n", rc);
}

ed_thread::ed_thread() {
    // entering here as main-thread
    int rc;
    pthread_attr_t attr;

    stat_run = 0; // initially ask user-thread to 'wait'
    pthread_cond_init  (&cond_run, NULL);
    pthread_mutex_init (&mutx_run, NULL);

    pthread_attr_init (&attr);
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_JOINABLE);

    rc = pthread_create (&id, &attr, (void*(*)(void*))(&Circular),(void *)this);
    if (rc) fprintf (stderr, "ed_thread: return code from pthread_create(id) is %d\n", rc);
    pthread_attr_destroy (&attr);
}

ed_thread::~ed_thread() {
    // entering here as main-thread
    int rc;
    rc = pthread_mutex_lock (&mutx_run);
    if (rc) fprintf (stderr, "~ed_thread: return code from pthread_mutex_lock(&mutx_run) is %d\n", rc);

    while (stat_run != 0) { // sleep as long as we (main-thread) haven't got 'wait' (i.e. 'completed') from user-thread
        rc = pthread_cond_wait (&cond_run, &mutx_run);
        if (rc) fprintf (stderr, "~ed_thread: return code from pthread_cond_wait(&cond_run) is %d\n", rc);
    }

    stat_run = (~0); // ('2') send 'terminate' to user-thread

    rc = pthread_cond_signal (&cond_run);
    if (rc) fprintf (stderr, "~ed_thread: return code from pthread_cond_signal(&cond_run) is %d\n", rc);

    rc = pthread_mutex_unlock (&mutx_run);
    if (rc) fprintf (stderr, "~ed_thread: return code from pthread_mutex_unlock(&mutx_run) is %d\n", rc);

    rc = pthread_join (id, /*status*/NULL); // wait here for 'thread termination' of user-thread
    if (rc) fprintf (stderr, "~ed_thread: return code from pthread_join(id) is %d\n", rc);

    rc = pthread_cond_destroy (&cond_run);
    if (rc) fprintf (stderr, "~ed_thread: return code from pthread_cond_destroy(&cond_run) is %d\n", rc);

    rc = pthread_mutex_destroy (&mutx_run);
    if (rc) fprintf (stderr, "~ed_thread: return code from pthread_mutex_destroy(&mutx_run) is %d\n", rc);
}

#endif /*WIN/PTHREADS*/
