/* Written by Dmitrij Schlesinger (2012),
 * www.inf.tu-dresden.de/~ds24/rtsegm/rtsegm.html
 *
 * See paper "A Real-time MRF Based Approach for Binary Segmentation"
 * at DAGM-OAGM 2012 Conference for details.
*/

// standard stuff
#include <stdio.h>
#include <iostream>
using namespace std;
#include <stdlib.h>
#include <time.h>

// input/output
#include "d_inout.h"

// rtsegm stuff
#include "rtsegm_gauss.h"
#include "rtsegm_segm.h"
#include "rtsegm_shape.h"

int main() {
    // Parameters:

    // Input -- camera, sizes etc.
    int mode=0;                 // input mode: 0 -- webcam, 1 -- video, 2 -- still image
    int camera=0;               // for mode=0
    const char *filename="..."; // for mode=1,2
    int height=240;	           // 0 means "use native (do not change)"
    int width=320;              // 0 means "use native (do not change)"
    bool flip=true;             // 0 -- do not flip, 1 -- mirror (for laptop webcams :-))
    int fps=0;                  // 0 means "try maximal" (helpful for webcams)
    int delay;		           // waiting time -- "almost" 1000/fps

    // modell (fixed parameter)
    // defaults are tuned for my working place:
    // quite light background, laptop webcam (30 fps), my face as the foreground,
    // quite fast computer (i7, quad-core, 2.8 GHz -- speed index about 2200)
    int ngauss=16;         // number of Gaussians
    double potts=1.3;	  // Potts parameter (>0, additive)
    double shapescale=2.5; // weight for the shape prior (>0, additive)

    // learning speed
    double stepgauss=0.9; // learning speed for Gaussians (0...1), the more the faster
    double stepmarg=0.1;	 // speed for marginals accumulation, 0.01 -- slow, 0.1 -- fast
    double stepshape=0.7; // learning speed for the shape (0...1), the more the faster

    // Input -- Webcam/Video/Image
    d_in *din=new d_in(mode,camera,(char *)filename,height,width,flip);
    height=din->height;
    width=din->width;

    if( fps ) delay=(int)(1000./fps+0.5);
    else delay=1;

    // array for the output image
    unsigned char *img_out=new unsigned char[width*height*3*4];

    // Output -- window
    d_out *dout=new d_out(0,(char *)"rtsegm",height*2,width*2,img_out);
    // Output -- video (no recording at the beginning)
    d_out *record=0;
    // Output -- screenshot
    d_out *screenshot;

    // Metropolis Sampler for Gaussians
    Rtsegm_gauss *gauss=new Rtsegm_gauss(height,width,ngauss,stepgauss);

    // Gibbs Sampler for Segmentations
    Rtsegm_segm *segm=new Rtsegm_segm(height,width,potts,stepmarg);

    // Shape optimizer
    Rtsegm_shape *shape=new Rtsegm_shape(height,width,shapescale,stepshape);

    // give access to the fields (see Fig.1 in the paper)
    // Rtsegm_gauss should observe the image
    gauss->img=din->img;
    // Rtsegm_gauss should observe the segmentation
    gauss->lab=segm->lab;
    // Rtsegm_segm should observe the field of Gaussian numbers
    segm->glab=gauss->glab;
    // Rtsegm_segm should observe the Gaussian weights
    segm->weights=gauss->weights;
    // Rtsegm_segm should observe the shape unary potentials
    segm->qsh=shape->qsh;
    // Rtsegm_shape should observe the segmentation
    shape->lab=segm->lab;

    // start
    gauss->start();
    segm->start();
    shape->start();

    // variables to watch real camera speed (frames per seconds)
    // and samplers/optimizer speeds (iterations per frame)
    int time0, frame_count;
    int countgprev, countgnext, countsprev, countsnext, countshprev, countshnext;
    double real_fps, gipf, sipf, shipf;

    time0=(int)time(0);
    frame_count=countgprev=countsprev=countshprev=0;
    if( fps ) real_fps=fps;
    else real_fps=10.;
    gipf=sipf=shipf=0.;

    // show some help
    cerr << endl << "     h -- help";
    cerr << endl << "     i -- info";
    cerr << endl << "     c -- start/stop capture";
    cerr << endl << "     s -- screenshot";
    cerr << endl << "     r -- reset";
    cerr << endl << "     p -- pause";
    cerr << endl << "   a/z -- potts +/-";
    cerr << endl << "   k/m -- shape +/-";
    cerr << endl << "   esc -- quit";
    cerr << endl << endl;

    // main loop
    bool flag_exit=false;
    while( !flag_exit ) {
        // wait a bit
        int key = din->key(delay);

        // watch real camera speed
        int time1=(int)time(0);
        if( time1>time0 ) {
            real_fps=real_fps*0.1+0.9*frame_count/(time1-time0);
            time0=time1;
            frame_count=0;
        }

        // watch sampler speed
        countgnext=gauss->it_count;
        countsnext=segm->it_count;
        countshnext=shape->it_count;
        gipf=gipf*0.99+0.01*(countgnext-countgprev);
        sipf=sipf*0.99+0.01*(countsnext-countsprev);
        shipf=shipf*0.99+0.01*(countshnext-countshprev);
        countgprev=countgnext;
        countsprev=countsnext;
        countshprev=countshnext;

        // what for a key
        if( key>0 ) {
            switch (key) {
            case 27: // escape -- exit
                flag_exit = true;
                break;
            case 'h': // show help
                cerr << endl << "     h -- help";
                cerr << endl << "     i -- info";
                cerr << endl << "     c -- start/stop capture";
                cerr << endl << "     s -- screenshot";
                cerr << endl << "     r -- reset";
                cerr << endl << "     p -- pause";
                cerr << endl << "   a/z -- potts +/-";
                cerr << endl << "   k/m -- shape +/-";
                cerr << endl << "   esc -- quit";
                cerr << endl << endl;
                break;
            case 'i': // show some info
                cerr << endl << height << "x" << width << ", " << real_fps << "fps, g:"
                     << gipf << "/" << gauss->step << ", s:"
                     << sipf << "/" << segm->step << " sh:"
                     << shipf << "/" << shape->step << ", speed index: "
                     << real_fps*(gipf+sipf+shipf);

                cerr << endl << "Potts: " << potts << ", shape: " << shapescale;

                cerr << endl << "Shape: x0=" << shape->x0 << ", y0="
                     << shape->y0 << ", rx=" << shape->rx << ", ry="
                     << shape->ry << ", level=" << shape->level;

                cerr << endl << "Sigma: " << gauss->sigma;

                cerr << endl << "Weights:" << endl;
                for( int g=0; g<ngauss; g++ )
                    cerr << " " << gauss->weights[g*2];
                cerr << endl;
                for( int g=0; g<ngauss; g++ )
                    cerr << " " << gauss->weights[g*2+1];
                cerr << endl << endl;
                break;
            case 'c': // start/stop recording
                if( record ) {
                    cerr << endl << "record finished" << endl;
                    delete record;
                    record=0;
                } else {
                    // Attention!!! The video is created with 30 fps -- the real speed is ignored
                    record=new d_out(1,(char *)"rtsegm.avi",height*2,width*2,img_out);
                    cerr << endl << "record started -- \"rtsegm.avi\"" << endl;
                }
                break;
            case 's': // take screenshot
                screenshot=new d_out(2,(char *)"rtsegm.png",height*2,width*2,img_out);
                screenshot->out();
                delete screenshot;
                cerr << endl << "screenshot done -- \"rtsegm.png\"" << endl;
                break;
            case 'r': // reset -- potts and shapescale are not affected
                segm->stop_me=true;
                gauss->stop_me=true;
                shape->stop_me=true;
                segm->wait();
                gauss->wait();
                shape->wait();
                segm->reset();
                gauss->reset();
                shape->reset();
                gauss->start();
                segm->start();
                shape->start();
                cerr << endl << "reset" << endl;
                break;
            case 'p': // pause
                segm->stop_me=true;
                gauss->stop_me=true;
                shape->stop_me=true;
                segm->wait();
                gauss->wait();
                shape->wait();
                cerr << endl << "pause, press any key to resume ... ";
                din->key(-1);
                gauss->start();
                segm->start();
                shape->start();
                cerr << "resumed" << endl;
                break;
            case 'a': // increase potts
                potts+=0.1;
                segm->potts=potts;
                cerr << endl << "potts=" << potts << endl;
                break;
            case 'z': // decrease potts
                potts-=0.1;
                if( potts<0.1 ) potts=0.1;
                segm->potts=potts;
                cerr << endl << "potts=" << potts << endl;
                break;
                break;
            case 'k':  // increase shapescale
                shapescale+=0.1;
                shape->scale=shapescale;
                cerr << endl << "shape=" << shapescale << endl;
                break;
            case 'm': // decrease shapescale
                shapescale-=0.1;
                if( shapescale<0.1 ) shapescale=0.1;
                shape->scale=shapescale;
                cerr << endl << "shape=" << shapescale << endl;
                break;
            default:
                break;
            }
        }

        // get the new frame
        if( !din->in() ) break; // 0 -- if there are no more frames in a video

        // prepare image for the visualization
        unsigned char *psrc=din->img;
        for( int ii=0, i=0; i<height; i++ ) {
            // input image -- left top
            unsigned char *pi=img_out+i*width*3*2;
            // Gaussians -- left bottom
            unsigned char *pg=pi+height*width*3*2;
            // labelling -- right top
            unsigned char *pl=pi+width*3;
            // marginals -- right bottom
            unsigned char *pm=pl+height*width*3*2;
            for( int j=0; j<width; j++,ii++ ) {
                // image
                *(pi++)=*(psrc++);
                *(pi++)=*(psrc++);
                *(pi++)=*(psrc++);
                // gaussians
                int ig=gauss->glab[ii];
                *(pg++)=(unsigned char)(gauss->m_r[ig]+0.5);
                *(pg++)=(unsigned char)(gauss->m_g[ig]+0.5);
                *(pg++)=(unsigned char)(gauss->m_b[ig]+0.5);
                // labelling
                unsigned char gray;
                gray=segm->lab[ii]*255;
                *(pl++)=gray;
                *(pl++)=gray;
                *(pl++)=gray;
                gray=(unsigned char)(segm->marg[ii]*255.+0.5);
                *(pm++)=gray;
                *(pm++)=gray;
                *(pm++)=gray;
            }
        }

        // draw ellipses
        int x0=(int)(shape->x0+0.5);
        int y0=(int)(shape->y0+0.5);
        int radx=(int)(shape->rx*1.41*sqrt(shape->level)+0.5);
        int rady=(int)(shape->ry*1.41*sqrt(shape->level)+0.5);
        dout->draw_ellipse(x0,y0,radx,rady);
        dout->draw_ellipse(x0+width,y0,radx,rady);
        dout->draw_ellipse(x0,y0+height,radx,rady);
        dout->draw_ellipse(x0+width,y0+height,radx,rady);

        // show in window
        dout->out();

        // record the frame (if necessary)
        if( record ) record->out();

        frame_count++;
    } // end of the main loop

    // stop recording (if any)
    if( record ) delete record;

    // stop all processes
    segm->stop_me=true;
    gauss->stop_me=true;
    shape->stop_me=true;
    segm->wait();
    gauss->wait();
    shape->wait();

    // clean
    delete gauss;
    delete segm;
    delete shape;
    delete din;
    delete dout;
    delete img_out;
    exit(0);
}
