/* Written by Dmitrij Schlesinger (2012),
 * www.inf.tu-dresden.de/~ds24/rtsegm/rtsegm.html
 *
 * See paper "A Real-time MRF Based Approach for Binary Segmentation"
 * at DAGM-OAGM 2012 Conference for details.
*/

#include <math.h>
#include <iostream>
using namespace std;
#include <stdlib.h>
#include <time.h>
#include "rtsegm_shape.h"

// constructor -- remember parameters, allocate arrays, initialize
Rtsegm_shape::Rtsegm_shape(int height, int width, double scale, double step)
    : ed_thread() {
    this->height=height;
    this->width=width;
    this->scale=scale;
    this->step=step;

    // init shape at the image center
    x0=width/2.;
    y0=height/2.;
    rx=ry=height/4.;
    level=1.;
    stop_me=false;
    it_count=0;
    qsh=new double[height*width];
    fill();
}

// destructor -- clean
Rtsegm_shape::~Rtsegm_shape() {
    delete qsh;
}

// the entry point for separate thread -- the shape optimizer
void Rtsegm_shape::run() {
    stop_me=false;

    // the main loop
    while( !stop_me ) {
        // zero statistics
        double xacc=0;
        double yacc=0;
        double rxacc=0;
        double ryacc=0;
        int n=0;
        // scan over the image and accumulate statistics
        int hw=height*width;
        for( int i=0; i<hw; i++ ) {
            if( lab[i] ) {
                int x=i%width;
                int y=i/width;
                yacc+=y;
                xacc+=x;
                double x2=x-x0;
                double y2=y-y0;
                rxacc+=x2*x2;
                ryacc+=y2*y2;
                n++;
            }
        }

        // update parameters
        if( n>10 ) { // to reduce instabilities
            x0=x0*(1.-step)+step*xacc/n;
            y0=y0*(1.-step)+step*yacc/n;
            rx=rx*(1.-step)+step*sqrt(2.*rxacc/n);
            ry=ry*(1.-step)+step*sqrt(2.*ryacc/n);
            level=level*(1.-step)+step*(n/(2.*3.1416*rx*ry));
        }

        // produce the unary potentials for the Gibbs Sampler
        fill();

        it_count++;
    }
    stop_me=false;
}

// calculates the unary potentials for Gibbs Sampling
void Rtsegm_shape::fill() {
    double sl=scale*level;
    double ry2=scale/(2.*ry*ry);
    double rx2=scale/(2.*rx*rx);
    for( int i=0, ii=0; i<height; i++ ) {
        for( int j=0; j<width; j++, ii++ ) {
            double i2=i-y0;
            double j2=j-x0;
            qsh[ii]=sl-i2*i2*ry2-j2*j2*rx2;
        }
    }
}

// reset the shape
void Rtsegm_shape::reset() {
    x0=width/2.;
    y0=height/2.;
    rx=ry=height/4.;
    level=1.;
    stop_me=false;
    fill();
}
